from gister.utils.text import wrap

class PrintLog():

  def info(message:str):
    print(f"\033[1;34;10m{wrap(message)}\033[0m\n")

  def success(message:str):
    print(f"\033[1;32;10m{wrap(message)} \U0001F44D\033[0m\n")

  def warning(message:str):
    print(f"\033[1;33;10m{wrap(message)} \U0001F9DF\033[0m\n")

  def error(message:str):
    print(f"\033[1;31;10m{wrap(message)} \U0001F525\033[0m\n")


from collections import namedtuple

from gister.errors import KeyInContextHasNoValueError, KeyNotInContextError
from gister.utils import asserts

ContextItemInfo = namedtuple('ContextItemInfo',
                             ['key',
                              'key_in_context',
                              'expected_type',
                              'is_expected_type',
                              'has_value'])

class Context(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pystring_globals = {}

    def __missing__(self, key):
        raise KeyNotInContextError(f"{key} not found in gister context.")

    def assert_child_key_has_value(self, parent, child, caller):
        asserts.assert_key_has_value(self, parent, caller)
        asserts.assert_key_has_value(self[parent], child, caller, parent)

    def assert_key_exists(self, key, caller):
        asserts.assert_key_exists(self, key, caller)

    def assert_key_has_value(self, key, caller):
        asserts.assert_key_has_value(self, key, caller)

    def assert_key_type_value(self, context_item, caller, extra_error_text=""):
        assert context_item, ("context_item parameter must be specified.")
        if extra_error_text is None or extra_error_text == '':
            append_error_text = ''
        else:
            append_error_text = f' {extra_error_text}'

        if not context_item["key_in_context"]:
            raise KeyNotInContextError(
                f'{caller} couldn\'t find '
                f'{context_item["key"]} in context.'
                f'{append_error_text}'
            )

        if not context_item["has_value"]:
            raise KeyInContextHasNoValueError(
                f'{caller} found {context_item["key"]} in '
                f'context but it doesn\'t have a value.'
                f'{append_error_text}'
            )
        if not context_item["is_expected_type"]:
            raise KeyInContextHasNoValueError(
                f'{caller} found {context_item["key"]} in context, but it\'s '
                f'not a {context_item["expected_type"]}.'
                f'{append_error_text}'
            )

    def assert_keys_exist(self, caller, *keys):
        assert keys, ("*keys parameter must be specified.")
        for key in keys:
            self.assert_key_exists(key, caller)

    def assert_keys_have_values(self, caller, *keys):
        for key in keys:
            self.assert_key_has_value(key, caller)

    def assert_child_keys_has_values(self, parent, child, keys, caller):
        for key in keys:
            self.assert_child_key_has_value(parent, key, caller)

    def assert_keys_type_value(self,
                               caller,
                               extra_error_text,
                               *context_items):

        assert context_items, ("context_items parameter must be specified.")

        for context_item in context_items:
            self.assert_key_type_value(context_item, caller, extra_error_text)


    def get_formatted(self, key):
        val = self[key]

        try:
            # any sort of complex type will work with recursive formatter.
            if val is None: raise KeyNotInContextError
            return val
        except KeyNotInContextError as err:
            # less cryptic error for end-user friendliness
            raise KeyNotInContextError(
                f'Unable to format \'{val}\' at context[\'{key}\'], '
                f'because {err}'
            ) from err



"""Cli entrypoint
"""
import argparse
from pathlib import Path
import argparse
import logging
import sys
import signal

import gister.runner
import gister.version
import gister.steps.gitenv
import gister.builder.command
import gister.log.logger
from gister.utils.text import wrap

log = logging.getLogger(__name__)

def get_parser():
    parser = argparse.ArgumentParser(allow_abbrev=True, prog='gister', description='Command line interface for the gister package',
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('--envm', help=wrap("Name of the enviroment. Should define user config " "inside ~/.gister"))
    parser.add_argument('--log', '--loglevel', dest='log_level', type=int,
                        default=None,
                        help=wrap(
                            'Integer log level. Defaults to 25 (NOTIFY).\n'
                            '10=DEBUG \n'
                            '20=INFO\n'
                            '25=NOTIFY\n'
                            '30=WARNING\n'
                            '40=ERROR\n'
                            '50=CRITICAL\n'
                            'Log Level < 10 gives full traceback on errors.'))
    parser.add_argument('--logpath', dest='log_path',
                        help=wrap(
                            'Log-file path. Append log output to this path.'))
    parser.add_argument('--version', action='version',
                        help='Version number', version=f'{gister.version.get_version()}')

    gister.builder.command.command_builder(parser)

    return parser


def get_args(args):
    # subparsers = parser.add_subparsers(help='Subcommands')
    # Parse all command line arguments
    return get_parser().parse_args(args)


""" Main
"""


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    parsed_args = get_args(args)

    try:
        gister.log.logger.set_root_logger(log_level=parsed_args.log_level, log_path=parsed_args.log_path)
        return gister.runner.main(
            git_env_change=parsed_args.env,
        )
    except KeyboardInterrupt:
        sys.stdout.write("\n")
        return 128 + signal.SIGINT
    except Exception as e:
        sys.stderr.write("\n")
        sys.stderr.write(f"\033[91m{type(e).__name__}: {str(e)}\033[0;0m")
        sys.stderr.write("\n")
        return 255


from pathlib import Path

from gister.errors import ContextError, FileTemplateError
from gister.log.print import PrintLog

""" Git template section
"""
def user_section(name="", email=""):
  return f'\
[user]\n\
  name="{name}"\n\
  email="{email}"'

def filter_section():
  return f'\n\
[filter "lfs"]\n\
  clean=git-lfs clean -- %f\n\
  smudge=git-lfs smudge -- %f\n\
  process=git-lfs filter-process\n\
  required=true\n'

def ssh_section():
  return f'\n\
[url "ssh://git@github.com/"]\n\
  insteadOf=https://github.com/\
  '


""" Process and Create template
"""
def process_template(template=""):
  try:
    if template == "":
      raise ContextError("Template must not empty")
    path = f"{Path.home()}/.gitconfig"
    with open(path, "w") as file:
      file.write(template)
      file.close()
      PrintLog.success("git environment switched!")

  except FileTemplateError:
    raise


def get_error_name(error):
    error_type = type(error)
    if error_type.__module__ in ['__main__', 'builtins']:
        return error_type.__name__
    else:
        return f'{error_type.__module__}.{error_type.__name__}'


class Error(Exception):
    """Base class for all exceptions"""

class ContextError(Error):
    """Error in context"""

class HandledError(Error):
    """Error that has already been saved to errors context collection"""

class KeyInContextHasNoValueError(ContextError):
    """context[key] doesn't have a value."""

class KeyNotInContextError(ContextError, KeyError):
    """Key not found in context."""

    def __str__(self):
        """Avoid KeyError custom error formatting."""
        return super(Exception, self).__str__()

class FileTemplateError(Error):
    """Error processing template"""


from collections.abc import MutableMapping
import logging
import ruamel.yaml as yaml
from pathlib import Path

from ruamel.yaml.main import YAML

logger = logging.getLogger(__name__)


def get_parsed_context():
    logger.debug("start reading file")

    # TODO: validate posible args
    # if not args:
    #     raise AssertionError(
    #         "Function must be invoked with context args set."
    #     )

    path = f"{Path.home()}/.gisterconfig"
    logger.debug("%d", )
    logger.debug("attempting to open file: %s", path)
    with open(path) as yaml_file:
        yaml_loader = yaml.YAML(typ='safe', pure=True)
        payload = yaml_loader.load(yaml_file)

        if not isinstance(payload, MutableMapping):
            raise TypeError("yaml file must be a dictionary at the top. Like key: value")

        logger.info("parsed. D: %d Done.", len(payload))
        return payload

def write_parsed_context(env):
    path = f"{Path.home()}/.gisterconfig"
    logger.debug("%d", )
    logger.debug("attempting to open file: %s", path)
    with open(path) as yaml_file:
        yaml_loader = yaml.YAML()
        payload = yaml_loader.load(yaml_file)

        payload["gitenv"]["active"] = env
        with open(path, "w") as fw:
            yaml_loader.dump(payload, fw)

class ConfigReader():

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__configpath = f"{Path.home()}/.gisterconfig"

    def read_yml_configs(self):
        path = self.__configpath
        logger.debug("%d", )
        logger.debug("attempting to open file: %s", path)
        with open(path) as yaml_file:
            yaml_loader = yaml.YAML()
            payload = yaml_loader.load(yaml_file)
            yaml_file.close()
        return payload

    def __writenewymlproperties__(self, payload):
        logger.debug("%d", )
        logger.debug("attempting to open file: %s", self.__configpath)
        with open(self.__configpath, "w") as fw:
            yaml_loader = yaml.YAML()
            yaml_loader.dump(payload, fw)
            fw.close()

    def read_json_configs(self):
        logger.debug("start reading file")

        # TODO: validate posible args
        # if not args:
        #     raise AssertionError(
        #         "Function must be invoked with context args set."
        #     )

        # path = f"{Path.home()}/.gisterconfig"
        path = self.__configpath
        logger.debug("%d", )
        logger.debug("attempting to open file: %s", path)
        with open(path) as yaml_file:
            yaml_loader = yaml.YAML(typ='safe', pure=True)
            payload = yaml_loader.load(yaml_file)

            if not isinstance(payload, MutableMapping):
                raise TypeError("yaml file must be a dictionary at the top. Like key: value")

            logger.info("parsed. D: %d Done.", len(payload))
            return payload
    def write_new_yml_properties(self, payload):
        self.__writenewymlproperties__(payload)


""" Commnad builder"""
from gister.log.print import PrintLog
def git_env(parser):
    subparser = parser.add_subparsers(help="Sub-commands")
    subparser_a = subparser.add_parser(
        'gitenv', help="Change git config environment")
    subparser_a.add_argument(
        '--env', help="Enviroment name", required=True)
    #subparser_a.add_argument('count', action='store', type=int)

switcher = {
    "gitenv": git_env,
}

""" Command builder
    Iterate hover switcher to initialize subcommands.
    To create new subcommands must define initializar function
    and integrate in switcher variable
"""
def command_builder(parser=None):
    # TODO: Validate subparser value and delete command argument
    for command in switcher:
        if switcher.get(command) is None:
            PrintLog.error("Command " f"{command} not found!")
            break
        func_exec = switcher.get(command)
        func_exec(parser)


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
# packages = ["gister"],
# url="https://github.com/pypa/sampleproject",
#project_urls={
#    "Bug Tracker": "https://github.com/pypa/sampleproject/issues",
#},
setup(
    name="gister-cli",
    version="0.0.2",
    author="Josh Andres",
    author_email="josh.an.bas@gmail.com",
    description="A awesome cli project.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=[
        'requests',
        'ruamel',
        'importlib; python_version == "3.6"',
    ],
    entry_points={
        'console_scripts': [
            'gister=gister.cli:main'
        ],
    },
)
